unit uCLI;
{$INCLUDE 'vfy.inc'}

interface

uses
  SysUtils, Classes, Windows,
  uConsoleApp, uCompatibility, uFiles,
  uHashItem, uVerify, uVerifyState;

type
  TDisplayingProgress = (dpNo, dpNeedCheckConsoleWidth, dpOnPreviousLine, dpOnNewLine);
  TVerifyApp = class(TConsoleApp)
    protected
      function AppTitle: string; override;
      function AppVersion: string; override;
      function DoExecute: integer; override;
      procedure DoConsoleCtrl(const CtrlType: DWORD; var Handled: boolean); override;
      procedure ShowSyntax(const Error: string = ''); override;
      procedure ParamInitialize; override;
      //procedure Log(const LogFileName: string; const Msg: string); override;
    private
      fVerify: TVerify;
      fControlEvent: DWORD;
      fDisplayingProgress: TDisplayingProgress;
      fLastProgressTicks: DWORD;
      fErrorLogFileName: string;
    protected
      function ParseCommandLine: integer; virtual;
      procedure VerifyStateChange(Sender: TVerify; State: TVerifyState; const Info: TStateChangeObject);
      procedure VerifyProgress(Sender: TVerify; const Position, Max: int64); virtual;
      procedure VerifyDeleteDuplicates(Sender: TVerify; FileNames: TStringList); virtual;
      procedure ClearProgress; virtual;
      procedure CheckControlEvent; virtual;
      property Verify: TVerify read fVerify;
      property ControlEvent: DWORD read fControlEvent write fControlEvent;
      property DisplayingProgress: TDisplayingProgress read fDisplayingProgress write fDisplayingProgress;
      property ErrorLogFileName: string read fErrorLogFileName write fErrorLogFileName;
    public
      constructor Create; override;
      destructor Destroy; override;
    end;

implementation

const
  PROGRESS_RESETCURSOR = #8#8#8#8#8#8#8;
  PROGRESS_WRITE = ' %3.3s.%d%%' + PROGRESS_RESETCURSOR;
  PROGRESS_CLEAR = '       ';
  PROGRESS_CLEAR_KEEP_POSITION = PROGRESS_CLEAR + PROGRESS_RESETCURSOR;

const
  PROGRESS_EVENT_TICKS = 1000;

const
  CTRL_NO_EVENT = $ffffffff;

const
  COLOR_BY_STATE: array[THashState] of TConsoleColor = (ccWhite, ccLightGreen, ccLightCyan, ccLightMagenta, ccYellow, ccYellow);

{ TVerifyApp }

constructor TVerifyApp.Create;
begin
  fVerify := TVerify.Create;
  fVerify.OnStateChange := VerifyStateChange;
  fVerify.OnProgress := VerifyProgress;
  fVerify.OnDeleteDuplicates := VerifyDeleteDuplicates;
  fControlEvent := CTRL_NO_EVENT;
  fDisplayingProgress := dpNo;
  inherited;
end;

destructor TVerifyApp.Destroy;
begin
  FreeAndNil(fVerify);
  inherited;
end;

function TVerifyApp.AppTitle: string;
begin
  Result := 'Verify' {$IFDEF WIN64} + '-x64' {$ENDIF} ;
end;

function TVerifyApp.AppVersion: string;
begin
  Result := {$INCLUDE 'vfy.version'} ;
end;

procedure TVerifyApp.ShowSyntax(const Error: string);
begin
  inherited;
  WriteColored(ccWhite, '[<arg>] [<arg>] ... <hashfile> [<dir>] [<dir>] ...'); Writeln;
  Writeln;
  Writeln('Commands:');
  WriteColored(ccWhite, ' -h, -?'); Writeln(' ......... Show this help screen.');
  Writeln;
  WriteColored(ccWhite, ' -b'); Writeln(' ............. Build a new hash list.');
  WriteColored(ccWhite, ' -v'); Writeln(' ............. Verify an existing hash list (default).');
  WriteColored(ccWhite, ' -u[+]'); Writeln(' .......... Verify and update an existing hash list. Use -u+ to allow a non-existent hash list.');
  WriteColored(ccWhite, ' -m <file>'); Writeln(' ...... Match new hash list <hashfile> against old <file>.');
  WriteColored(ccWhite, ' -D <file>'); Writeln(' ...... Find files from <hashfile> which exist in <file>.');
  WriteColored(ccWhite, ' -D1'); Writeln(' ............ Find duplicate files within <hashfile>.');
  WriteColored(ccWhite, ' -DD'); Writeln(' ............ Delete duplicate files. <hashfile> may be a results file, too. The operation is interactive!');
  WriteColored(ccWhite, ' -s'); Writeln(' ............. Sort hash list <hashfile>. A sorted hashlist is faster to read.');
  WriteColored(ccWhite, ' -T'); Writeln(' ............. Test hash list <hashfile>. Verifies that the hashlist is fully readable.');
  Writeln;
  Writeln('Matching specifiers:'); 
  WriteColored(ccWhite, ' -j[-]'); Writeln(' .......... [Don''t] Scan junctions and reparse points (default: -j).');
  WriteColored(ccWhite, ' -r[-]'); Writeln(' .......... [Don''t] Recursively scan subdirectories (default: -r).');
  WriteColored(ccWhite, ' -t[-]'); Writeln(' .......... [Don''t] Use file modification timestamp (default: -t).');
  WriteColored(ccWhite, ' -tp <msec>'); Writeln(' ..... Use <msec> millisecond precision for comparison (default: 0).');
  WriteColored(ccWhite, ' -c <file>'); Writeln(' ...... Use <file> as a source for getting file information. That is, if file X is to be read from the disk, try to read it from <file> first.');
  {$IFDEF CASESENSITIVESUPPORT}
  WriteColored(ccWhite, ' -cs[-]'); Writeln(' ......... [Don''t] Use case-sensitive file functions (default:: -cs-).');
  {$ENDIF}
  {$IFDEF OVERRIDEWOW64}
  WriteColored(ccWhite, ' -wow64[-]'); Writeln(' ...... [Disable] Wow64 filesystem redirection (default: -wow64).');
  {$ENDIF}
  {$IFDEF RAWFILEACCESS}
  WriteColored(ccWhite, ' -raw[-]'); Writeln(' ........ [Disable] support for raw filesystem access (default: -raw-).');
  {$ENDIF}
  WriteColored(ccWhite, ' -pf <file>'); Writeln(' ..... Use <file> as a progress cache (read if exists, write during processing)');
  Writeln;
  Writeln('Source files:'); 
  WriteColored(ccWhite, ' -d <dir>'); Writeln(' ....... Add directory <dir> to scanned directories.');
  WriteColored(ccWhite, ' -rd <dir>'); Writeln(' ...... Do not write the initial <dir> to output hashlist.');
  WriteColored(ccWhite, ' -fd[-]'); Writeln(' ......... [Don''t] Write full directory name to output hashlist (needed for cache generation).');
  WriteColored(ccWhite, ' -i <wildcard>'); Writeln('... Include files whose name matches <wildcard>.');
  WriteColored(ccWhite, ' -x <wildcard>'); Writeln('... Exclude files whose name matches <wildcard>.');
  WriteColored(ccWhite, ' -ir <regexp>'); Writeln('.... Include files whose full name matches <regexp>.');
  WriteColored(ccWhite, ' -xr <regexp>'); Writeln('.... Exclude files whose full name matches <regexp>.');
  Writeln;
  Writeln('Output settings:'); 
  WriteColored(ccWhite, ' -p[-]'); Writeln(' .......... [Don''t] Show progress.');
  WriteColored(ccWhite, ' -e <file>'); Writeln(' ...... Write error log to file.');
  WriteColored(ccWhite, ' -o <file>'); Writeln(' ...... Write verify/match/dupes results to <file>.');
  WriteColored(ccWhite, ' -sm[-]'); Writeln(' ......... [Don''t] Show matched (non-changed) files (default: -sm-).');
  WriteColored(ccWhite, ' -sa[-]'); Writeln(' ......... [Don''t] Show added (new) files (default: -sa).');
  WriteColored(ccWhite, ' -sc[-]'); Writeln(' ......... [Don''t] Show changed files (default: -sc).');
  WriteColored(ccWhite, ' -sd[-]'); Writeln(' ......... [Don''t] Show deleted files (default: -sd).');
  WriteColored(ccWhite, ' -sp[-]'); Writeln(' ......... [Don''t] Show copied files (default: -sp).');
  WriteColored(ccWhite, ' -sr[-]'); Writeln(' ......... [Don''t] Show renamed files (default: -sr).');
  {$IFDEF DEBUG_OPERATIONSPROGRESS}
  WriteColored(ccWhite, ' --debug <file>'); Writeln(' . Create a debugging log.');
  {$ENDIF}
  Writeln;
  {$IFDEF DELPHICZBANNER}
  Writeln('This utility was entered in a Delphi.cz competition. You can read the details'); 
  Writeln('of this competition here:'); 
  WriteColored(ccLightCyan, 'http://delphi.cz/post/Soutez-o-RAD-Studio-2007-Architekt.aspx'); Writeln;
  Writeln;
  {$ENDIF}
  //Writeln;
end;

procedure TVerifyApp.ParamInitialize;
begin
  inherited;
  Verify.Clear;
  if StdOutRedirected then
    Verify.ShowProgress := False;
  ErrorLogFileName := '';
end;

function TVerifyApp.DoExecute: integer;
begin
  // Read command line arguments
  ParamInitialize;
  Result := ParseCommandLine;
  if Result <> EXITCODE_OK then
    Exit;
  try
    ControlEvent := CTRL_NO_EVENT;
    ShowProgressBar(pbsUnknown);
    try
      ClearProgress;
      case Verify.VerifyCommand of
        vcBuild:
          Result := Verify.Build;
        vcVerify, vcDefault:
          Result := Verify.Verify;
        vcUpdate:
          Result := Verify.Update;
        vcMatch:
          Result := Verify.Match;
        vcDupes:
          Result := Verify.Dupes(False);
        vcDupes1:
          Result := Verify.Dupes(True);
        vcDeleteDupes:
          Result := Verify.DeleteDupes;
        vcSort:
          Result := Verify.Sort;
        vcTest:
          Result := Verify.Test;
        else
          Result := EXITCODE_UNKNOWN_COMMAND;
        end;
    finally
      ShowProgressBar(pbsNoProgress);
    end;
  except
    on E: EAbort do
      Result := EXITCODE_BREAK;
    on E: Exception do begin
      if ErrorLogFileName <> '' then
        Log(ErrorLogFileName, #13#10'ERROR %s: %s', [E.ClassName, E.Message]);
      Raise;
    end;
  end;
end;

function TVerifyApp.ParseCommandLine: integer;
const ONLY_ONE_COMMAND = 'Only one of -b/-v/-u/-m/-D/-D1/-s commands may be present.';

  procedure SyntaxError(const Msg: string); overload;
    begin
      ShowSyntax(Msg);
      Abort;
    end;
  procedure SyntaxError(const Msg: string; const Params: array of const); overload;
    begin
      SyntaxError(Format(Msg, Params));
    end;

var Param: string;
    iParam: integer;
begin
  try
    while ParamGetNext(Param) do
      begin
      if Param[1] = '-' then
        begin
        if (Param = '-h') or (Param = '-?') then
          SyntaxError('')
        else if Param = '-v' then
          if Verify.VerifyCommand = vcDefault then
            Verify.VerifyCommand := vcVerify
          else
            SyntaxError(ONLY_ONE_COMMAND)
        else if Param = '-b' then
          if Verify.VerifyCommand = vcDefault then
            Verify.VerifyCommand := vcBuild
          else
            SyntaxError(ONLY_ONE_COMMAND)
        else if Param = '-u' then
          if Verify.VerifyCommand = vcDefault then
            Verify.VerifyCommand := vcUpdate
          else
            SyntaxError(ONLY_ONE_COMMAND)
        else if Param = '-u+' then
          if Verify.VerifyCommand = vcDefault then
          begin
            Verify.VerifyCommand := vcUpdate;
            Verify.HashFileMustExist := False;
          end
          else
            SyntaxError(ONLY_ONE_COMMAND)
        else if Param = '-m' then
          if Verify.VerifyCommand = vcDefault then
            if ParamGetNext(Param) then
              if FileExists(Param) then
                begin
                Verify.VerifyCommand := vcMatch;
                Verify.MatchListFileName := Param;
                end
              else
                SyntaxError('Matchfile not found for -m command.')
            else
              SyntaxError('File name must be provided for -m command.')
          else
            SyntaxError(ONLY_ONE_COMMAND)
        else if Param = '-D' then
          if Verify.VerifyCommand = vcDefault then
            if ParamGetNext(Param) then
              if FileExists(Param) then
                begin
                Verify.VerifyCommand := vcDupes;
                Verify.MatchListFileName := Param;
                end
              else
                SyntaxError('Dupesfile not found for -D command.')
            else
              SyntaxError('File name must be provided for -D command.')
          else
            SyntaxError(ONLY_ONE_COMMAND)
        else if Param = '-D1' then
          if Verify.VerifyCommand = vcDefault then
            Verify.VerifyCommand := vcDupes1
          else
            SyntaxError(ONLY_ONE_COMMAND)
        else if Param = '-DD' then
          if Verify.VerifyCommand = vcDefault then
            Verify.VerifyCommand := vcDeleteDupes
          else
            SyntaxError(ONLY_ONE_COMMAND)
        else if Param = '-s' then
          if Verify.VerifyCommand = vcDefault then
            Verify.VerifyCommand := vcSort
          else
            SyntaxError(ONLY_ONE_COMMAND)
        else if Param = '-T' then
          if Verify.VerifyCommand = vcDefault then
            Verify.VerifyCommand := vcTest
          else
            SyntaxError(ONLY_ONE_COMMAND)
        else if Param = '-r' then
          Verify.Recursive := True
        else if Param = '-r-' then
          Verify.Recursive := False
        else if Param = '-p' then
          Verify.ShowProgress := True
        else if Param = '-p-' then
          Verify.ShowProgress := False
        else if Param = '-j' then
          Verify.Junctions := True
        else if Param = '-j-' then
          Verify.Junctions := False
        else if Param = '-t' then
          Verify.UseTimestamp := True
        else if Param = '-t-' then
          Verify.UseTimestamp := False
        else if Param = '-tp' then
          if ParamGetNext(Param) then
            begin
            iParam := StrToIntDef(Param, -1);
            if iParam >= 0 then
              Verify.TimestampPrecision := iParam
            else
              SyntaxError('Precision must be an integer number, not "%s".', [Param]);
            end
          else
            SyntaxError('Precision must be provided with -tp argument.')
        {$IFDEF CASESENSITIVESUPPORT}
        else if Param = '-cs' then
          Verify.CaseSensitive := True
        else if Param = '-cs-' then
          Verify.CaseSensitive := False
        {$ENDIF}
        {$IFDEF OVERRIDEWOW64}
        else if Param = '-wow64' then
          Verify.DisableWow64 := False
        else if Param = '-wow64-' then
          Verify.DisableWow64 := True
        {$ENDIF}
        {$IFDEF RAWFILEACCESS}
        else if Param = '-raw' then
          Verify.RawFileAccess := True
        else if Param = '-raw-' then
          Verify.RawFileAccess := False
        {$ENDIF}
        else if Param = '-sm' then
          Verify.HideStates := Verify.HideStates - [hsOK]
        else if Param = '-sm-' then
          Verify.HideStates := Verify.HideStates + [hsOK]
        else if Param = '-sa' then
          Verify.HideStates := Verify.HideStates - [hsNew]
        else if Param = '-sa-' then
          Verify.HideStates := Verify.HideStates + [hsNew]
        else if Param = '-sc' then
          Verify.HideStates := Verify.HideStates - [hsChanged]
        else if Param = '-sc-' then
          Verify.HideStates := Verify.HideStates + [hsChanged]
        else if Param = '-sd' then
          Verify.HideStates := Verify.HideStates - [hsDeleted]
        else if Param = '-sd-' then
          Verify.HideStates := Verify.HideStates + [hsDeleted]
        else if Param = '-sp' then
          Verify.HideStates := Verify.HideStates - [hsCopied]
        else if Param = '-sp-' then
          Verify.HideStates := Verify.HideStates + [hsCopied]
        else if Param = '-sr' then
          Verify.HideStates := Verify.HideStates - [hsRenamed]
        else if Param = '-sr-' then
          Verify.HideStates := Verify.HideStates + [hsRenamed]
        else if Param = '-i' then
          if ParamGetNext(Param) then
            Verify.AddIncludeExcludeMask(Param, iemInclude)
          else
            SyntaxError('Mask must be provided with -i argument.')
        else if Param = '-x' then
          if ParamGetNext(Param) then
            Verify.AddIncludeExcludeMask(Param, iemExclude)
          else
            SyntaxError('Mask must be provided with -x argument.')
        else if Param = '-ir' then
          if ParamGetNext(Param) then
            Verify.AddIncludeExcludeMask(Param, iemIncludeRegExp)
          else
            SyntaxError('Regular expression must be provided with -ir argument.')
        else if Param = '-xr' then
          if ParamGetNext(Param) then
            Verify.AddIncludeExcludeMask(Param, iemExcludeRegExp)
          else
            SyntaxError('Regular expression ask must be provided with -xr argument.')
        else if Param = '-d' then
          if ParamGetNext(Param) then
            Verify.BaseDirectories.Add(Param)
          else
            SyntaxError('Directory name must be provided with -d argument.')
        else if Param = '-rd' then
          if ParamGetNext(Param) then
            Verify.BaseDirectoriesRoot := Param
          else
            SyntaxError('Directory name must be provided with -rd argument.')
        else if Param = '-fd' then
          Verify.SaveFullPaths := True
        else if Param = '-fd-' then
          Verify.SaveFullPaths := False
        else if Param = '-o' then
          if ParamGetNext(Param) then
            if Param = '-' then
              Verify.OutputFileName := ''
            else
              Verify.OutputFileName := Param
          else
            SyntaxError('File name must be provided with -o argument.')
        else if Param = '-pf' then
          if ParamGetNext(Param) then
            if Param = '-' then
              Verify.ProgressFileName := ''
            else
              Verify.ProgressFileName := Param
          else
            SyntaxError('File name must be provided with -pf argument.')
        else if Param = '-c' then
          if ParamGetNext(Param) then
            if Param = '-' then
              Verify.CacheFileName := ''
            else
              Verify.CacheFileName := Param
          else
            SyntaxError('File name must be provided with -o argument.')
        else if Param = '-e' then
          if ParamGetNext(Param) then
            ErrorLogFileName := Param
          else
            SyntaxError('File name must be provided with -e argument.')
        {$IFDEF DEBUG_OPERATIONSPROGRESS}
        else if Param = '--debug' then
          if ParamGetNext(Param) then
            Verify.LogFileName := Param
          else
            SyntaxError('File name must be provided with --debug argument.')
        {$ENDIF}
        else
          SyntaxError('Unknown argument "%s".', [Param])
        end
      else
        if Verify.HashesFileName = '' then
          Verify.HashesFileName := Param
        else
          Verify.BaseDirectories.Add(Param);
      end;
    if Verify.HashesFileName = '' then
      SyntaxError('Hashfile must be provided.');
    if Verify.BaseDirectories.Count = 0 then
      Verify.BaseDirectories.Add('.');
    Result := EXITCODE_OK;
  except
    on EAbort do
      Result := EXITCODE_SYNTAX;
    end;
end;

procedure TVerifyApp.VerifyStateChange(Sender: TVerify; State: TVerifyState; const Info: TStateChangeObject);

  procedure WriteVerifyStat(const Title: string; const Value: integer; const ZeroColor, NonZeroColor: TConsoleColor);
    begin
      Write(Title);
      if Value = 0 then
        WriteColored(ZeroColor, '0')
      else
        WriteColored(NonZeroColor, IntToStr(Value));
    end;

var
  Msg: string;
  HashItem: THashItem;
begin
  case State of
    vstFileError:
      begin
        with Info as TStateChangeErrorInFile do
          Msg := Format(#13#10'ERROR %s for file "%s": %s', [ExceptionClass, FileName, ExceptionMessage]);
        if ErrorLogFileName <> '' then
          Log(ErrorLogFileName, Msg);
        ClearProgress;
        ShowError(Msg);
      end;
    vstLoadingHashFile:
      begin
        Msg := (Info as TStateChangeFilename).FileName;
        ClearProgress;
        Write('Loading hash file '); WriteColored(ccWhite, Msg); Write(' ... ');
        DisplayingProgress := dpNeedCheckConsoleWidth;
      end;
    vstBuildOutput:
      ; // Nothing to do
    vstMatchOutput:
      if Verify.OutputFileName = '' then
        begin
          HashItem := (Info as TStateChangeHashItem).Item;
          if not (HashItem.State in Verify.HideStates) then
          begin
            ClearProgress;
            WritelnColored(COLOR_BY_STATE[HashItem.State], HashItem.MatchOutput(False));
          end;
        end;
    vstDupesOutput:
      if Verify.OutputFileName = '' then
        begin
          HashItem := (Info as TStateChangeHashItem).Item;
          ClearProgress;
          WritelnColored(COLOR_BY_STATE[HashItem.State], HashItem.DupesOutput(False));
        end;
    vstProgressDone:
      begin
        ClearProgress;
      end;
    vstSortingResults:
      begin
        ClearProgress;
        Writeln('Sorting results...');
      end;
    vstWritingResults:
      begin
        ClearProgress;
        Writeln('Writing results...');
      end;
    vstSortingHashlist:
      begin
        ClearProgress;
        Writeln('Sorting hash list...');
      end;
    vstComparingHashlists:
      begin
        ClearProgress;
        Write('Comparing hash lists...');
        DisplayingProgress := dpNeedCheckConsoleWidth;
      end;
    vstLookingForDupes:
      begin
        ClearProgress;
        Write('Looking for dupes...');
        DisplayingProgress := dpNeedCheckConsoleWidth;
      end;
    vstProcessingFile:
      begin
        Msg := AnsiQuotedStr((Info as TStateChangeFilename).FileName, '"');
        ClearProgress;
        Write(Msg);
        DisplayingProgress := dpNeedCheckConsoleWidth;
      end;
    vstMatchStats,
    vstDupesStats:
      begin
        ClearProgress;
        Writeln('Operation statistics:');
        with Info as TStateChangeStatistics do
          begin
          WriteVerifyStat('Matched: ', MatchCount,   ccWhite, COLOR_BY_STATE[hsOK]); Write('    ');
          WriteVerifyStat('New: ',     NewCount,     ccWhite, COLOR_BY_STATE[hsNew]); Write('    ');
          WriteVerifyStat('Changed: ', ChangedCount, ccWhite, COLOR_BY_STATE[hsChanged]); Write('    ');
          WriteVerifyStat('Deleted: ', DeletedCount, ccWhite, COLOR_BY_STATE[hsDeleted]); Write('    ');
          WriteVerifyStat('Copied: ',  CopiedCount,  ccWhite, COLOR_BY_STATE[hsCopied]); Write('    ');
          WriteVerifyStat('Renamed: ', RenamedCount, ccWhite, COLOR_BY_STATE[hsRenamed]); Writeln;
          end;
        Writeln;
      end;
  end;
end;

procedure TVerifyApp.VerifyDeleteDuplicates(Sender: TVerify; FileNames: TStringList);
var
  FileName, FileToKeepStr: string;
  i, FileToKeep: integer;
  Again: boolean;
begin
  ClearProgress;
  repeat
    Again := False;
    if Verify.WasAborted then
      Break;
    FileNames.CaseSensitive := {$IFDEF CASESENSITIVESUPPORT} Verify.CaseSensitive {$ELSE} False {$ENDIF} ;
    FileNames.Sort;
    for i := Pred(FileNames.Count) downto 0 do
    begin
      FileName := FileNames[i];
      if (FileName = '') or (not FileExists(FileName)) then
        FileNames.Delete(i);
    end;
    if FileNames.Count <= 1 then
      Break;
    Writeln;
    WritelnColored(ccWhite, Format('Found %d duplicates:', [FileNames.Count]));
    for i := 0 to Pred(FileNames.Count) do
    begin
      WriteColored(ccLightCyan, Format('%6.6s', [IntToStr(Succ(i))]));
      Write(') ');
      Writeln(FileNames[i]);
    end;
    WriteColored(ccWhite, 'Which file do you want to keep? [all] ');
    Readln(FileToKeepStr);
    if FileToKeepStr = '' then
      Break;
    FileToKeep := StrToIntDef(FileToKeepStr, -1);
    if FileToKeep = 0 then
      Break;
    if (FileToKeep > 0) and (FileToKeep <= FileNames.Count) then
    begin
      for i := 0 to Pred(FileNames.Count) do
      begin
        if i <> Pred(FileToKeep) then
        begin
          FileName := FileNames[i];
          if FileExists(FileName) then
          begin
            if not SysUtils.DeleteFile(FileName) then
            begin
              ShowError('Failed to delete: %s', [FileName]);
              Again := True;
            end;
          end;
        end;
      end;
    end;
  until not Again;
end;

procedure TVerifyApp.VerifyProgress(Sender: TVerify; const Position, Max: int64);
var Ticks: DWORD;
    PromileDone: int64;
    Width, Height, CursorX, CursorY: integer;
begin
  CheckControlEvent;
  if not Verify.ShowProgress then
    Exit;
  if StdOutRedirected then
    Exit;
  if DisplayingProgress in [dpNo, dpNeedCheckConsoleWidth] then
  begin
    if not GetConsoleSize(Width, Height, CursorX, CursorY) then
      Exit;
    if (CursorX + Length(PROGRESS_CLEAR) + 1) >= Width then
    begin
      DisplayingProgress := dpOnNewLine;
      Writeln;
    end
    else
    begin
      DisplayingProgress := dpOnPreviousLine;
    end;
  end;
  Ticks := GetTickCount;
  if (Position = 0) or ((Ticks - fLastProgressTicks) >= PROGRESS_EVENT_TICKS) then
  begin
    fLastProgressTicks := Ticks;
    PromileDone := 1000 * Position div Max;
    Write(Format(PROGRESS_WRITE, [IntToStr(PromileDone div 10), PromileDone mod 10]));
  end;
end;

procedure TVerifyApp.DoConsoleCtrl(const CtrlType: DWORD; var Handled: boolean);
begin
  if ControlEvent = CTRL_NO_EVENT then // Regular processing on double Ctrl+Break
    if CtrlType in [CTRL_C_EVENT, CTRL_BREAK_EVENT, CTRL_CLOSE_EVENT] then
      begin
      ControlEvent := CtrlType;
      Handled := True;
      end;
  inherited;
end;

procedure TVerifyApp.CheckControlEvent;
var Msg, Response: string;
begin
  case ControlEvent of
    CTRL_C_EVENT:
      Msg := 'Ctrl+C pressed.';
    CTRL_BREAK_EVENT:
      Msg := 'Ctrl+Break pressed.';
    CTRL_CLOSE_EVENT:
      Msg := 'Application close requested.';
    else
      Msg := '';
    end;
  if Msg <> '' then
    if Verify <> nil then
      begin
      ControlEvent := CTRL_NO_EVENT;
      ClearProgress;
      ShowError(Msg);
      ShowProgressBar(pbsPaused);
      try
        if Verify.ProgressFileName <> '' then
          Response := 'Q'
        else
        begin
          WriteColored(ccWhite, 'Do you want to quit ('); WriteColored(ccLightRed, 'Q');
          WriteColored(ccWhite, ') now or show results first (anything else)?');
          Readln(Response);
        end;
      finally
        ShowProgressBar(pbsUnknown);
      end;
      if (Response <> 'Q') and (Response <> 'q') then
        Verify.DoAbort
      else
        Abort;
      end;
end;

procedure TVerifyApp.ClearProgress;
begin
  CheckControlEvent;
  //dpNo, dpNeedCheckConsoleWidth, dpOnPreviousLine, dpOnNewLine
  if DisplayingProgress <> dpNo then
  begin
    if DisplayingProgress = dpOnNewLine then
      Write(PROGRESS_CLEAR_KEEP_POSITION)
    else
      Writeln(PROGRESS_CLEAR);
    DisplayingProgress := dpNo;
  end;
end;

end.
