unit uVerify;
{$INCLUDE 'vfy.inc'}

interface

uses
  SysUtils, Classes, Windows,
  uCompatibility, uSystem, uFiles, uPCRE,
  {$IFDEF RAWFILEACCESSUSING7ZIP} SevenZip, {$ENDIF}
  uHashList, uHashItem, uVerifyState;

type
  TVerify = class;

  TProcessFileFunction = function (const BasePath, Path: string; Tag: integer): integer of object;
  TAfterProcessFunction = procedure (var ProcessResult: integer; Tag: integer) of object;

  TIncludeExcludeMask = (iemInclude, iemExclude, iemIncludeRegExp, iemExcludeRegExp);
  TVerifyCommand = (vcDefault, vcVerify, vcBuild, vcUpdate, vcMatch, vcDupes, vcDupes1, vcDeleteDupes, vcSort, vcTest);

  TStateChangeEvent = procedure(Sender: TVerify; State: TVerifyState; const Info: TStateChangeObject) of object;
  TProgressEvent = procedure(Sender: TVerify; const Position, Max: int64) of object;
  TDeleteDuplicatesEvent = procedure(Sender: TVerify; FileNames: TStringList) of object;

  TVerify = class(TObject)
    private
      fVerifyCommand: TVerifyCommand;
      fHashesFileName: string;
      fHashFileMustExist: boolean;
      fOutputFileName: string;
      fCacheFileName: string;
      fMatchListFileName: string;
      fBaseDirectories: TStringList;
      fRecursive: boolean;
      fJunctions: boolean;
      fHashList: THashList;
      fAuxHashList: THashList;
      fCache: THashList;
      fShowProgress: boolean;
      fIncludeExcludeList: TStringList;
      {$IFDEF CASESENSITIVESUPPORT}
      fCaseSensitive: boolean;
      {$ENDIF}
      {$IFDEF OVERRIDEWOW64}
      fWow64RedirectionOldValueDONOTCHANGE: Pointer;
      fDisableWow64: boolean;
      {$ENDIF}
      {$IFDEF RAWFILEACCESS}
      fRawFileAccess: boolean;
      {$ENDIF}
      fUseTimestamp: boolean;
      fTimestampPrecision: integer;
      fHideStates: THashStates;
      fOnStateChange: TStateChangeEvent;
      fOnProgress: TProgressEvent;
      fAborted: boolean;
      fHashFileStream: TTextStream;
      fBaseDirectoriesRoot: string;
      fSaveFullPaths: boolean;
      {$IFDEF DEBUG_LOG}
      fLogFileName: string;
      fLogStream: TTextStream;
      {$ENDIF}
      fProgressFileName: string;
      fProgressStream: TTextStream;
      fOnDeleteDuplicates: TDeleteDuplicatesEvent;
      {$IFDEF DEBUG_LOG}
      procedure SetLogFileName(const Value: string);
      {$ENDIF}
    protected
      procedure DoStateChange(State: TVerifyState; Info: TStateChangeObject = nil; FreeInfo: boolean = True); virtual;
      procedure DoProgress(const Position, Max: int64); virtual;
      function FixUpDirectory(const Dir: string): string; virtual;
      procedure FixUpBaseDirectories; virtual;
      procedure SeparateBaseDir(const Dir: string; out BaseDir, Path: string);
      procedure HashingProgress(Sender: TObject; Done, Total: int64); virtual;
      function ErrorInFile(E: Exception; const BasePath, Path: string): integer; virtual;
      procedure DeleteOutputFile;
      function ProcessFiles(const BasePath, Path: string; ProcessFileFunction: TProcessFileFunction; Tag: integer): integer; virtual;
      function ProcessFile_Build(const BasePath, Path: string; Tag: integer): integer; virtual;
      function ProcessFile_Verify(const BasePath, Path: string; Tag: integer): integer; virtual;
      function ProcessFile_Update(const BasePath, Path: string; Tag: integer): integer; virtual;
      procedure AfterProcess_Verify(var ProcessResult: integer; Tag: integer); virtual;
      procedure AfterProcess_Update(var ProcessResult: integer; Tag: integer); virtual;
      function InternalVerify(ProcessFileFunction: TProcessFileFunction; AfterProcessFunction: TAfterProcessFunction; Tag: integer): integer; virtual;
      function InternalMatch: integer; virtual;
      procedure ReadHashItem(HashItem: THashItem; const BasePath, Path: string);
      procedure ApplyIncludeExcludeToHashList(HashList: THashList); virtual;
      function IsFileWanted(const Path: string): boolean; virtual;
      procedure CheckControlEvent; virtual;
      procedure LoadHashFile(const FileName: string; HashList: THashList; ClearFirst: boolean = True; ChangesFile: boolean = False); virtual;
      procedure LoadHashOrChangesFile(const FileName: string; HashList: THashList; ClearFirst: boolean = True); virtual;
      {$IFDEF CASESENSITIVESUPPORT}
      function FindFirst(const Path: string; Attr: Integer; var SR: TSearchRec): Integer; virtual;
      {$ENDIF}
      {$IFDEF OVERRIDEWOW64}
        procedure ChangeWow64Redirection(Active: boolean); virtual;
        {$IFNDEF OVERRIDEWOW64GLOBALLY}
          {$IFDEF DELPHI6_UP}
            {$MESSAGE ERROR 'OVERRIDEWOW64 is not done yet.'}
          {$ELSE}
            'OVERRIDEWOW64 is not done yet.'
          {$ENDIF}
        {$ENDIF}
      {$ENDIF}
      {$IFDEF RAWFILEACCESS}
        procedure ChangeRawFileAccessStatus(Active: boolean); virtual;
      {$ENDIF}
      procedure SetupProgressStream;
      procedure ClearProgressFile;
      procedure WriteToProgressStream(Item: THashItem);
      property ProgressStream: TTextStream read fProgressStream;
      property HashFileStream: TTextStream read fHashFileStream;
      property AuxHashList: THashList read fAuxHashList;
      property Cache: THashList read fCache;
      property IncludeExcludeList: TStringList read fIncludeExcludeList;
      property Aborted: boolean read fAborted write fAborted;
      {$IFDEF DEBUG_LOG}
      procedure DebugLog(const Msg: string); overload;
      procedure DebugLog(const Msg: string; const Params: array of const); overload;
      property LogStream: TTextStream read fLogStream;
      {$ENDIF}
    public
      constructor Create; virtual;
      destructor Destroy; override;
      procedure Clear; virtual;
      procedure DoAbort; virtual;
      function WasAborted: boolean; virtual;
      function Build: integer; virtual;
      function Verify: integer; virtual;
      function Update: integer; virtual;
      function Match: integer; virtual;
      function Dupes(OneFileOnly: boolean): integer; virtual;
      function DeleteDupes: integer; virtual;
      function Sort: integer; virtual;
      function Test: integer; virtual;
      procedure ClearIncludeExcludeMasks; virtual;
      procedure AddIncludeExcludeMask(const Mask: string; MaskType: TIncludeExcludeMask); virtual;
      property VerifyCommand: TVerifyCommand read fVerifyCommand write fVerifyCommand;
      property HashesFileName: string read fHashesFileName write fHashesFileName;
      property HashFileMustExist: boolean read fHashFileMustExist write fHashFileMustExist;
      property OutputFileName: string read fOutputFileName write fOutputFileName;
      property CacheFileName: string read fCacheFileName write fCacheFileName;
      property ProgressFileName: string read fProgressFileName write fProgressFileName;
      property MatchListFileName: string read fMatchListFileName write fMatchListFileName;
      property BaseDirectories: TStringList read fBaseDirectories;
      property BaseDirectoriesRoot: string read fBaseDirectoriesRoot write fBaseDirectoriesRoot;
      property Recursive: boolean read fRecursive write fRecursive;
      property ShowProgress: boolean read fShowProgress write fShowProgress;
      property Junctions: boolean read fJunctions write fJunctions;
      {$IFDEF CASESENSITIVESUPPORT}
      property CaseSensitive: boolean read fCaseSensitive write fCaseSensitive;
      {$ENDIF}
      {$IFDEF OVERRIDEWOW64}
      property DisableWow64: boolean read fDisableWow64 write fDisableWow64;
      {$ENDIF}
      {$IFDEF RAWFILEACCESS}
      property RawFileAccess: boolean read fRawFileAccess write fRawFileAccess;
      {$ENDIF}
      property SaveFullPaths: boolean read fSaveFullPaths write fSaveFullPaths;
      property UseTimestamp: boolean read fUseTimestamp write fUseTimestamp;
      property TimestampPrecision: integer read fTimestampPrecision write fTimestampPrecision; // Milliseconds
      property HideStates: THashStates read fHideStates write fHideStates;
      property OnStateChange: TStateChangeEvent read fOnStateChange write fOnStateChange;
      property OnProgress: TProgressEvent read fOnProgress write fOnProgress;
      property HashList: THashList read fHashList;
      {$IFDEF DEBUG_LOG}
      property LogFileName: string read fLogFileName write SetLogFileName;
      {$ENDIF}
      property OnDeleteDuplicates: TDeleteDuplicatesEvent read fOnDeleteDuplicates write fOnDeleteDuplicates;
    end;

const
  EXITCODE_OK = 0;
  EXITCODE_CHANGED_FILES = 1;
  EXITCODE_UNKNOWN_COMMAND = 101;
  EXITCODE_NO_DATA = 102;
  EXITCODE_ERROR_IN_FILES = 103;
  EXITCODE_BREAK = 253;
  EXITCODE_SYNTAX = 254;

implementation

{$IFDEF OVERRIDEWOW64GLOBALLY}
  {$IFDEF DELPHI6_UP}
    {$MESSAGE HINT 'Using OVERRIDEWOW64GLOBALLY. Don''t use LoadLibrary during Build/Verify/Update!'}
  {$ENDIF}
{$ENDIF}

const
  INCEXCTYPE_INCLUDE = 'I';
  INCEXCTYPE_EXCLUDE = 'X';

const
  TAG_BUILD = 1;
  TAG_VERIFY = 2;
  TAG_VERIFYFAST = 3;
  TAG_UPDATE = 4;
  TAG_UPDATEFAST = 5;

type
  EDelayedAbort = class(EAbort);

{ TVerify }

constructor TVerify.Create;
begin
  fBaseDirectories := TStringList.Create;
  fHashList := THashList.Create;
  fAuxHashList := THashList.Create;
  fCache := THashList.Create;
  fIncludeExcludeList := TStringList.Create;
  {$IFDEF OVERRIDEWOW64}
  fWow64RedirectionOldValueDONOTCHANGE := nil;
  {$ENDIF}
  {$IFDEF DEBUG_LOG}
  fLogStream := nil;
  {$ENDIF}
  inherited;
end;

procedure TVerify.DeleteOutputFile;
begin
  {$IFDEF DEBUG_OPERATIONSPROGRESS}
  DebugLog('TVerify.DeleteOutputFile: filename "%s"', [OutputFileName]);
  {$ENDIF}
  if OutputFileName <> '' then
    if FileExists(OutputFileName) then
      SysUtils.DeleteFile(OutputFileName);
end;

destructor TVerify.Destroy;
begin
  Clear;
  FreeAndNil(fBaseDirectories);
  FreeAndNil(fHashList);
  FreeAndNil(fCache);
  FreeAndNil(fAuxHashList);
  FreeAndNil(fIncludeExcludeList);
  FreeAndNil(fProgressStream);
  {$IFDEF DEBUG_LOG}
  FreeAndNil(fLogStream);
  {$ENDIF}
  inherited;
end;

procedure TVerify.Clear;
begin
  HashList.Clear;
  Cache.Clear;
  FreeAndNil(fProgressStream);
  ClearIncludeExcludeMasks;
  VerifyCommand := vcDefault;
  HashesFileName := '';
  HashFileMustExist := True;
  OutputFileName := '';
  ProgressFileName := '';
  MatchListFileName := '';
  BaseDirectories.Clear;
  BaseDirectoriesRoot := '';
  Recursive := True;
  Junctions := True;
  ShowProgress := True;
  {$IFDEF CASESENSITIVESUPPORT}
  CaseSensitive := False;
  {$ENDIF}
  {$IFDEF OVERRIDEWOW64}
  DisableWow64 := False;
  {$ENDIF}
  {$IFDEF RAWFILEACCESS}
  RawFileAccess := False;
  {$ENDIF}
  UseTimestamp := True;
  TimestampPrecision := 0;
  HideStates := [hsOK];
  Aborted := False;
end;

procedure TVerify.DoAbort;
begin
  Aborted := True;
end;

procedure TVerify.DoStateChange(State: TVerifyState; Info: TStateChangeObject; FreeInfo: boolean);
begin
  try
    if Assigned(OnStateChange) then
      OnStateChange(Self, State, Info);
  finally
    if Info <> nil then
      if FreeInfo then
        FreeAndNil(Info);
  end;
end;

procedure TVerify.DoProgress(const Position, Max: int64);
begin
  if Assigned(OnProgress) then
    OnProgress(Self, Position, Max);
end;

procedure TVerify.FixUpBaseDirectories;
var
  i: integer;
begin
  for i := 0 to Pred(BaseDirectories.Count) do
    BaseDirectories[i] := FixUpDirectory(BaseDirectories[i]);
end;

function TVerify.FixUpDirectory(const Dir: string): string;
begin
  if Dir = '' then
    Result := IncludeTrailingPathDelimiter(GetCurrentDir)
  else
    Result := IncludeTrailingPathDelimiter(ExpandFileName(Dir));
  {$IFDEF LONGPATHS}
  if Result <> '' then
    Result := GetVeryLongFileName(Result);
  {$ENDIF}
end;

function TVerify.ErrorInFile(E: Exception; const BasePath, Path: string): integer;
begin
  {$IFDEF DEBUG_OPERATIONSPROGRESS}
  DebugLog('TVerify.ErrorInFile(%s."%s", "%s", "%s"): start', [E.ClassName, E.Message, BasePath, Path]);
  {$ENDIF}
  if E is EAbort then
    Result := EXITCODE_BREAK
  else
  begin
    Result := EXITCODE_ERROR_IN_FILES;
    DoStateChange(vstFileError, TStateChangeErrorInFile.Create(E, BasePath + Path));
  end;
end;

procedure TVerify.LoadHashFile(const FileName: string; HashList: THashList; ClearFirst, ChangesFile: boolean);
begin
  {$IFDEF DEBUG_OPERATIONSPROGRESS}
  DebugLog('TVerify.LoadHashFile("%s"): start', [FileName]);
  {$ENDIF}
  if ClearFirst then
    HashList.Clear;
  if ShowProgress then
    DoStateChange(vstLoadingHashFile, TStateChangeFilename.Create(FileName));
  if ChangesFile then
    HashList.LoadFromChangesFile(FileName, '', HashingProgress)
  else
    HashList.LoadFromFile(FileName, '', HashingProgress);
  ApplyIncludeExcludeToHashList(HashList);
  if ShowProgress then
    DoStateChange(vstProgressDone);
  {$IFDEF DEBUG_OPERATIONSPROGRESS}
  DebugLog('TVerify.LoadHashFile("%s"): done', [FileName]);
  {$ENDIF}
end;

procedure TVerify.LoadHashOrChangesFile(const FileName: string; HashList: THashList; ClearFirst: boolean = True);
begin
  {$IFDEF DEBUG_OPERATIONSPROGRESS}
  DebugLog('TVerify.LoadHashOrChangesFile("%s"): start', [FileName]);
  {$ENDIF}
  if ClearFirst then
    HashList.Clear;
  if ShowProgress then
    DoStateChange(vstLoadingHashFile, TStateChangeFilename.Create(FileName));
  HashList.LoadFromRegularOrChangesFile(FileName, '', HashingProgress);
  ApplyIncludeExcludeToHashList(HashList);
  if ShowProgress then
    DoStateChange(vstProgressDone);
  {$IFDEF DEBUG_OPERATIONSPROGRESS}
  DebugLog('TVerify.LoadHashOrChangesFile("%s"): done', [FileName]);
  {$ENDIF}
end;

function TVerify.Build: integer;
var
  i, Code: integer;
  BaseDir, Path: string;
begin
  {$IFDEF DEBUG_OPERATIONSPROGRESS}
  DebugLog('TVerify.Build: start');
  {$ENDIF}
  Result := EXITCODE_NO_DATA;
  HashList.Clear;
  AuxHashList.Clear;
  Cache.Clear;
  if CacheFileName <> '' then
  begin
    {$IFDEF DEBUG_OPERATIONSPROGRESS}
    DebugLog('TVerify.Build: loading cache');
    {$ENDIF}
    LoadHashFile(CacheFileName, Cache, False);
  end;
  if (ProgressFileName <> '') and (FileExists(ProgressFileName)) then
  begin
    {$IFDEF DEBUG_OPERATIONSPROGRESS}
    DebugLog('TVerify.Build: loading progress file');
    {$ENDIF}
    LoadHashFile(ProgressFileName, Cache, False, True);
  end;
  if Cache.Count > 0 then
    Cache.SortByPath;
  {$IFDEF DEBUG_OPERATIONSPROGRESS}
  DebugLog('TVerify.Build: deleting output file');
  {$ENDIF}
  DeleteOutputFile;
  {$IFDEF OVERRIDEWOW64GLOBALLY}
  if DisableWow64 then
    ChangeWow64Redirection(False);
  {$ENDIF}
  try
    {$IFDEF RAWFILEACCESS}
    ChangeRawFileAccessStatus(RawFileAccess);
    try
    {$ENDIF}
    {$IFDEF DEBUG_OPERATIONSPROGRESS}
    DebugLog('TVerify.Build: creating hashfile "%s"', [HashesFileName]);
    {$ENDIF}
    fHashFileStream := TTextStream.Create(HashesFileName, fmCreate, fmShareDenyWrite);
    try
      if FileExists(HashesFileName) then
        SysUtils.DeleteFile(HashesFileName);
      {$IFDEF DEBUG_OPERATIONSPROGRESS}
      DebugLog('TVerify.Build: processing files');
      {$ENDIF}
      FixUpBaseDirectories;
      try
        SetupProgressStream;
        try
          for i := 0 to Pred(BaseDirectories.Count) do
          begin
            SeparateBaseDir(BaseDirectories[i], BaseDir, Path);
            Code := ProcessFiles(BaseDir, Path, ProcessFile_Build, TAG_BUILD);
            if Result = EXITCODE_NO_DATA then
              Result := Code
            else if Code <> EXITCODE_OK then
              Result := Code;
          end;
        finally
          FreeAndNil(fProgressStream);
        end;
      except
        on E: EDelayedAbort do
          Result := EXITCODE_BREAK
      end;
      {$IFDEF DEBUG_OPERATIONSPROGRESS}
      DebugLog('TVerify.Build: processing files complete');
      {$ENDIF}
      FreeAndNil(fHashFileStream);
      if Result <> EXITCODE_BREAK then
      begin
        if ShowProgress then
          DoStateChange(vstSortingResults);
        {$IFDEF DEBUG_OPERATIONSPROGRESS}
        DebugLog('TVerify.Build: sorting hashfile');
        {$ENDIF}
        HashList.SortByPath;
        {$IFDEF DEBUG_OPERATIONSPROGRESS}
        DebugLog('TVerify.Build: saving hashfile "%s"', [HashesFileName]);
        {$ENDIF}
        HashList.SaveToFile(HashesFileName);
        {$IFDEF DEBUG_OPERATIONSPROGRESS}
        DebugLog('TVerify.Build: saving hashfile complete');
        {$ENDIF}
        ClearProgressFile;
      end;
    finally
      FreeAndNil(fHashFileStream);
    end;
    {$IFDEF RAWFILEACCESS}
    finally
      ChangeRawFileAccessStatus(False);
    end;
    {$ENDIF}
  finally
    {$IFDEF OVERRIDEWOW64GLOBALLY}
    if DisableWow64 then
      ChangeWow64Redirection(True);
    {$ENDIF}
  end;
  {$IFDEF DEBUG_OPERATIONSPROGRESS}
  DebugLog('TVerify.Build: done');
  {$ENDIF}
end;

function TVerify.ProcessFile_Build(const BasePath, Path: string; Tag: integer): integer;
var
  HashItem: THashItem;
begin
  {$IFDEF DEBUG_OPERATIONSPROGRESS}
  DebugLog('TVerify.ProcessFile_Build("%s", "%s"): start', [BasePath, Path]);
  {$ENDIF}
  Result := EXITCODE_OK;
  try
    HashItem := THashItem.Create;
    try
      ReadHashItem(HashItem, BasePath, Path);
      HashItem.State := hsNew;
      HashList.Add(HashItem);
      WriteToProgressStream(HashItem);
      {$IFDEF SHOWRESULTSIMMEDIATELY}
      HashFileStream.WriteLine(HashItem.BuildOutput(True));
      {$ENDIF}
      if ShowProgress then
        DoStateChange(vstBuildOutput, TStateChangeHashItem.Create(HashItem));
    except
      FreeAndNil(HashItem);
      Raise;
    end;
  except
    on E: EAbort do
      Raise;
    on E: Exception do
      Result := ErrorInFile(E, BasePath, Path);
  end;
  {$IFDEF DEBUG_OPERATIONSPROGRESS}
  DebugLog('TVerify.ProcessFile_Build("%s", "%s"): done', [BasePath, Path]);
  {$ENDIF}
end;

function TVerify.InternalVerify(ProcessFileFunction: TProcessFileFunction; AfterProcessFunction: TAfterProcessFunction; Tag: integer): integer;
var
  i, Code: integer;
  BaseDir, Path: string;
  HashItem: THashItem;
begin
  {$IFDEF DEBUG_OPERATIONSPROGRESS}
  DebugLog('TVerify.InternalVerify: start');
  {$ENDIF}
  Result := EXITCODE_NO_DATA;
  HashList.Clear;
  AuxHashList.Clear;
  Cache.Clear;
  if CacheFileName <> '' then
  begin
    {$IFDEF DEBUG_OPERATIONSPROGRESS}
    DebugLog('TVerify.InternalVerify: loading cache');
    {$ENDIF}
    LoadHashFile(CacheFileName, Cache, False);
  end;
  if (ProgressFileName <> '') and (FileExists(ProgressFileName)) then
  begin
    {$IFDEF DEBUG_OPERATIONSPROGRESS}
    DebugLog('TVerify.InternalVerify: loading progress file');
    {$ENDIF}
    LoadHashFile(ProgressFileName, Cache, False, True);
  end;
  if Cache.Count > 0 then
    Cache.SortByPath;
  {$IFDEF DEBUG_OPERATIONSPROGRESS}
  DebugLog('TVerify.InternalVerify: deleting output file');
  {$ENDIF}
  DeleteOutputFile;
  {$IFDEF OVERRIDEWOW64GLOBALLY}
  if DisableWow64 then
    ChangeWow64Redirection(False);
  try
  {$ENDIF}
  {$IFDEF RAWFILEACCESS}
  ChangeRawFileAccessStatus(RawFileAccess);
  try
  {$ENDIF}
  FixUpBaseDirectories;
  {$IFDEF DEBUG_OPERATIONSPROGRESS}
  DebugLog('TVerify.InternalVerify: loading hashfile');
  {$ENDIF}
  if HashFileMustExist or FileExists(HashesFileName) then
  begin
    LoadHashFile(HashesFileName, HashList);
    if ShowProgress then
      DoStateChange(vstSortingHashlist);
    HashList.SortByPath;
  end;
  for i := 0 to Pred(HashList.Count) do
  begin
    HashItem := HashList[i];
    HashItem.State := hsOK;
    AuxHashList.AddCopy(HashItem);
    HashItem.State := hsDeleted;
  end;
  try
    {$IFDEF DEBUG_OPERATIONSPROGRESS}
    DebugLog('TVerify.InternalVerify: processing files');
    {$ENDIF}
    SetupProgressStream;
    try
      for i := 0 to Pred(BaseDirectories.Count) do
      begin
        SeparateBaseDir(BaseDirectories[i], BaseDir, Path);
        Code := ProcessFiles(BaseDir, Path, ProcessFileFunction, Tag);
        if Result = EXITCODE_NO_DATA then
          Result := Code
        else if Code <> EXITCODE_OK then
          Result := Code;
      end;
      {$IFDEF DEBUG_OPERATIONSPROGRESS}
      DebugLog('TVerify.InternalVerify: processing files complete');
      {$ENDIF}
    finally
      FreeAndNil(fProgressStream);
    end;
  except
    on E: EDelayedAbort do
      Result := EXITCODE_BREAK;
  end;
  if Assigned(AfterProcessFunction) then
    AfterProcessFunction(Result, Tag);
  {$IFDEF DEBUG_OPERATIONSPROGRESS}
  DebugLog('TVerify.InternalVerify: determining differences');
  {$ENDIF}
  Code := InternalMatch;
  if (Result = EXITCODE_OK) or (Result = EXITCODE_CHANGED_FILES) then
    Result := Code;
  ClearProgressFile;
  {$IFDEF RAWFILEACCESS}
  finally
    ChangeRawFileAccessStatus(False);
  end;
  {$ENDIF}
  {$IFDEF OVERRIDEWOW64GLOBALLY}
  finally
    if DisableWow64 then
      ChangeWow64Redirection(True);
  end;
  {$ENDIF}
  {$IFDEF DEBUG_OPERATIONSPROGRESS}
  DebugLog('TVerify.InternalVerify: done');
  {$ENDIF}
end;

function TVerify.Verify: integer;
begin
  {$IFDEF DEBUG_OPERATIONSPROGRESS}
  DebugLog('TVerify.Verify: start');
  {$ENDIF}
  Result := InternalVerify(ProcessFile_Verify, AfterProcess_Verify, TAG_VERIFY);
  {$IFDEF DEBUG_OPERATIONSPROGRESS}
  DebugLog('TVerify.Verify: done');
  {$ENDIF}
end;

function TVerify.WasAborted: boolean;
begin
  Result := Aborted;
end;

procedure TVerify.WriteToProgressStream(Item: THashItem);
var
  OldPath: string;
begin
  if ProgressStream <> nil then
  begin
    OldPath := Item.Path;
    Item.Path := GetRealFileName(Item.BasePath + Item.Path);
    try
      ProgressStream.WriteLine(Item.MatchOutput(True));
      ProgressStream.Flush;
    finally
      Item.Path := OldPath;
    end;
  end;
end;

function TVerify.ProcessFile_Verify(const BasePath, Path: string; Tag: integer): integer;
const
  MILLISECONDS_PER_DAY = 24 * 60 * 60 * 1000;
var
  HashItem: THashItem;
  HashListFileName: string;
  ExistsInHashFile: boolean;
  Index: integer;
  {$IFDEF DEBUG_REFRESHNEEDED}
  PreviousFileName, NextFileName: string;
  {$ENDIF}
begin
  {$IFDEF DEBUG_OPERATIONSPROGRESS}
  DebugLog('TVerify.ProcessFile_Verify("%s", "%s"): start', [BasePath, Path]);
  {$ENDIF}
  Result := EXITCODE_OK;
  try
    if SaveFullPaths then
      HashListFileName := GetRealFileName(BasePath + Path)
    else
      HashListFileName := Path;
    ExistsInHashFile := HashList.FindByPath(HashlistFileName, Index);
    if UseTimeStamp then
    begin
      if ExistsInHashFile then
      begin
        if (Abs(HashList[Index].Date - THashItem.GetFileDate(BasePath + Path)) <= (TimestampPrecision * 1/MILLISECONDS_PER_DAY)) then
        begin
          if (HashList[Index].Size = THashItem.GetFileSize(BasePath + Path)) then
          begin
            HashList[Index].Path := HashListFileName;
            HashList[Index].State := hsOK;
            WriteToProgressStream(HashList[Index]);
            Result := EXITCODE_OK;
            {$IFDEF DEBUG_OPERATIONSPROGRESS}
            DebugLog('TVerify.ProcessFile_Verify("%s", "%s"): done', [BasePath, Path]);
            {$ENDIF}
            Exit;
          {$IFDEF DEBUG_REFRESHNEEDED}
          end
          else
          begin
            DebugLog('TVerify.ProcessFile_Verify("%s"): sizes differ: real=%d, stored=%d', [HashListFileName, THashItem.GetFileSize(BasePath + Path), HashList[Index].Size]);
          {$ENDIF}
          end;
        {$IFDEF DEBUG_REFRESHNEEDED}
        end
        else
        begin
          DebugLog('TVerify.ProcessFile_Verify("%s"): dates differ: real="%s", stored="%s"', [HashListFileName, FormatDateTime('yyyy"-"mm"-"dd" "hh":"nn":"ss"."zzz', THashItem.GetFileDate(BasePath + Path)), FormatDateTime('yyyy"-"mm"-"dd" "hh":"nn":"ss"."zzz', HashList[Index].Date)]);
        {$ENDIF}
        end;
      {$IFDEF DEBUG_REFRESHNEEDED}
      end
      else
      begin
        if (Index >= 0) and (Index < HashList.Count) then
          NextFileName := HashList[Index].FileName
        else
          NextFileName := '';
        if (Index >= 1) then
          PreviousFileName := HashList[Index-1].FileName
        else
          PreviousFileName := '';
        DebugLog('TVerify.ProcessFile_Verify("%s"): stored file not found at index %d; previous="%s", next="%s"', [HashListFileName, Index, PreviousFileName, NextFileName]);
      {$ENDIF}
      end;
    end;
    HashItem := THashItem.Create;
    try
      ReadHashItem(HashItem, BasePath, Path);
      if ExistsInHashFile then
        if (HashList[Index].Size = HashItem.Size) and (HashList[Index].Hash = HashItem.Hash) then
        begin
          HashList[Index].Assign(HashItem);
          HashList[Index].State := hsOK;
          WriteToProgressStream(HashList[Index]);
          FreeAndNil(HashItem);
        end
        else
        begin
          HashItem.State := hsChanged;
          //HashList[Index].Free; // Free will be done by THashList itself due to OwnObjects=True
          HashList[Index] := HashItem;
          WriteToProgressStream(HashItem);
          HashItem := nil;
          Result := EXITCODE_CHANGED_FILES;
        end
      else
      begin
        HashItem.State := hsNew;
        HashList.Insert(Index, HashItem);
        WriteToProgressStream(HashItem);
        HashItem := nil;
        Result := EXITCODE_CHANGED_FILES;
      end;
    except
      FreeAndNil(HashItem);
      Raise;
    end;
  except
    on E: EAbort do
      Raise;
    on E: Exception do
      Result := ErrorInFile(E, BasePath, Path);
  end;
  {$IFDEF DEBUG_OPERATIONSPROGRESS}
  DebugLog('TVerify.ProcessFile_Verify("%s", "%s"): done', [BasePath, Path]);
  {$ENDIF}
end;

procedure TVerify.ReadHashItem(HashItem: THashItem; const BasePath, Path: string);
var
  CacheItem: THashItem;
  CachedFileName: string;
  Index: integer;
begin
  if SaveFullPaths or ((Cache <> nil) and (Cache.Count > 0)) then
    CachedFileName := GetRealFileName(BasePath + Path)
  else
    CachedFileName := '';
  CacheItem := nil;
  if (Cache <> nil) and (Cache.Count > 0) and (CachedFileName <> '') and Cache.FindByPath(CachedFileName, Index) then
  begin
    CacheItem := Cache[Index];
    if (CacheItem.GetFileDate(CachedFileName) <> CacheItem.Date) or (CacheItem.GetFileSize(CachedFileName) <> CacheItem.Size) then
  {$IFDEF DEBUG_CACHE}
    begin
      DebugLog('TVerify.ReadHashItem("%s" -> "%s"): in cache, dirty', [Path, CachedFileName]);
  {$ENDIF}
      CacheItem := nil;
  {$IFDEF DEBUG_CACHE}
    end
    else
    begin
      if CachedFileName <> '' then
        DebugLog('TVerify.ReadHashItem("%s" -> "%s"): in cache, up to date', [Path, CachedFileName]);
    end;
  end
  else
  begin
    if CachedFileName <> '' then
      DebugLog('TVerify.ReadHashItem("%s" -> "%s"): not in cache', [Path, CachedFileName]);
  {$ENDIF}
  end;
  if CacheItem <> nil then
  begin
    HashItem.Assign(CacheItem);
    HashItem.BasePath := BasePath;
    HashItem.Path := Path;
  end
  else
  begin
    HashItem.ReadFromFile(BasePath, Path, HashingProgress {$IFDEF CASESENSITIVESUPPORT} , CaseSensitive {$ENDIF} );
  end;
  if SaveFullPaths and (CachedFileName <> '') then
  begin
    HashItem.BasePath := '';
    HashItem.Path := CachedFileName;
  end;
end;

procedure TVerify.SeparateBaseDir(const Dir: string; out BaseDir, Path: string);
var
  ix, RootLen: integer;
  Root: string;
begin
  ix := Pos('|', Dir);
  if ix > 0 then
  begin
    BaseDir := IncludeTrailingPathDelimiter(Copy(Dir, 1, Pred(ix)));
    if ix < Length(Dir) then
      if Dir[Succ(ix)] = '\' then
        Inc(ix);
    Path := Copy(Dir, Succ(ix), MaxInt);
  end
  else
  begin
    BaseDir := Dir;
    Path := '';
    if BaseDirectoriesRoot <> '' then
    begin
      Root := FixUpDirectory(BaseDirectoriesRoot);
      RootLen := Length(Root);
      if AnsiCompareText(Root, Copy(Dir, 1, Length(Root))) = 0 then
      begin
        BaseDir := IncludeTrailingPathDelimiter(Root);
        if RootLen < Length(Dir) then
          if Dir[Succ(RootLen)] = '\' then
            Inc(RootLen);
        Path := Copy(Dir, Succ(RootLen), MaxInt);
      end;
    end;
  end;
end;

procedure TVerify.SetupProgressStream;
begin
  FreeAndNil(fProgressStream);
  if ProgressFileName <> '' then
  begin
    if FileExists(ProgressFileName) then
    begin
      fProgressStream := TTextStream.Create(ProgressFileName, fmOpenReadWrite, fmShareDenyWrite);
      fProgressStream.Position := fProgressStream.Size;
    end
    else
    begin
      fProgressStream := TTextStream.Create(ProgressFileName, fmCreate, fmShareDenyWrite);
    end;
  end;
end;

procedure TVerify.AfterProcess_Verify(var ProcessResult: integer; Tag: integer);
begin
  {$IFDEF DEBUG_OPERATIONSPROGRESS}
  DebugLog('TVerify.AfterProcess_Verify: start');
  {$ENDIF}
  HashList.DeleteByState(hsDeleted);
  {$IFDEF DEBUG_OPERATIONSPROGRESS}
  DebugLog('TVerify.AfterProcess_Verify: done');
  {$ENDIF}
end;

function TVerify.Update: integer;
begin
  {$IFDEF DEBUG_OPERATIONSPROGRESS}
  DebugLog('TVerify.Update: start');
  {$ENDIF}
  Result := InternalVerify(ProcessFile_Update, AfterProcess_Update, TAG_UPDATE);
  {$IFDEF DEBUG_OPERATIONSPROGRESS}
  DebugLog('TVerify.Update: done');
  {$ENDIF}
end;

function TVerify.ProcessFile_Update(const BasePath, Path: string; Tag: integer): integer;
begin
  Result := ProcessFile_Verify(BasePath, Path, Tag);
end;

procedure TVerify.AfterProcess_Update(var ProcessResult: integer; Tag: integer);
begin
  {$IFDEF DEBUG_OPERATIONSPROGRESS}
  DebugLog('TVerify.AfterProcess_Update: start');
  {$ENDIF}
  AfterProcess_Verify(ProcessResult, Tag);
  HashList.SortByPath;
  {$IFDEF DEBUG_OPERATIONSPROGRESS}
  DebugLog('TVerify.AfterProcess_Update: saving hashfile "%s"', [HashesFileName]);
  {$ENDIF}
  HashList.SaveToFile(HashesFileName);
  {$IFDEF DEBUG_OPERATIONSPROGRESS}
  DebugLog('TVerify.AfterProcess_Update: done');
  {$ENDIF}
end;

function TVerify.InternalMatch: integer;
  // HashList - new (current) hashlist
  // AuxHashList - old hashlist
var
  i, Index: integer;
  Done, Total: int64;
  State: THashState;
  StateCount: array[THashState] of integer;
  OutputStream: TTextStream;

  procedure Step;
    begin
      Inc(Done);
      HashingProgress(Self, Done, Total);
    end;

begin
  {$IFDEF DEBUG_OPERATIONSPROGRESS}
  DebugLog('TVerify.InternalMatch: start');
  {$ENDIF}
  Result := EXITCODE_OK;
  if ShowProgress then
    DoStateChange(vstComparingHashlists);
  Total := 2 * AuxHashList.Count + 2 * HashList.Count;
  Done := -1;
  Step;
  for i := 0 to Pred(AuxHashList.Count) do
  begin
    AuxHashList[i].State := hsDeleted;
    Step;
  end;
  // 1. Match by path
  {$IFDEF DEBUG_OPERATIONSPROGRESS}
  DebugLog('TVerify.InternalMatch: match by path');
  {$ENDIF}
  AuxHashList.SortByPath;
  for i := 0 to Pred(HashList.Count) do
  begin
    if AuxHashList.FindByPath(HashList[i].Path, Index) then
    begin
      HashList[i].State := hsChanged;
      HashList[i].Related := AuxHashList[Index];
      while (Index < AuxHashList.Count) and (HashItemCompareByPath(HashList[i], AuxHashList[Index]) = 0) do
      begin
        if HashItemCompareByHash(HashList[i], AuxHashList[Index]) = 0 then
        begin
          if HashList[i].State <> hsOK then
          begin
            if HashList[i].Path = AuxHashList[Index].Path then
            begin
              HashList[i].State := hsOK;
            end
            else
            begin
              HashList[i].State := hsRenamed;
            end;
            HashList[i].Related := AuxHashList[Index];
          end;
          AuxHashList[Index].State := hsOK;
        end
        else
        begin
          if AuxHashList[Index].State = hsDeleted then
            AuxHashList[Index].State := hsChanged;
        end;
        Inc(Index);
      end;
    end
    else
    begin
      HashList[i].Related := nil;
      HashList[i].State := hsNew;
    end;
    Step;
    end;
  // 2. Match not-matched by hash
  {$IFDEF DEBUG_OPERATIONSPROGRESS}
  DebugLog('TVerify.InternalMatch: match by hash');
  {$ENDIF}
  AuxHashList.SortByHash;
  for i := 0 to Pred(HashList.Count) do
  begin
    if not (HashList[i].State in [hsOK, hsChanged]) then
      if AuxHashList.FindByHash(HashList[i].Hash, HashList[i].Size, Index) then
      begin
        if HashList[i].Related = AuxHashList[Index] then
          HashList[i].State := hsRenamed
        else
          HashList[i].State := hsCopied;
        HashList[i].Related := AuxHashList[Index];
        while (Index < AuxHashList.Count) and (HashItemCompareByHash(HashList[i], AuxHashList[Index]) = 0) do
        begin
          if AuxHashList[Index].State <> hsOK then
          begin
            AuxHashList[Index].State := hsRenamed;
            if (AuxHashList[Index].Related = nil) or (AuxHashList[Index] <> AuxHashList[Index].Related.Related) then
            begin
              HashList[i].Related := AuxHashList[Index];
              AuxHashList[Index].Related := HashList[i];
            end;
          end;
          Inc(Index);
        end;
      end;
      Step;
    end;
  // 3. Process missing or renamed files from the old hashes
  {$IFDEF DEBUG_OPERATIONSPROGRESS}
  DebugLog('TVerify.InternalMatch: process missing/renamed');
  {$ENDIF}
  for i := 0 to Pred(AuxHashList.Count) do
  begin
    if AuxHashList[i].State = hsRenamed then
      if AuxHashList[i].Related <> nil then
        if AuxHashList[i].Related.State = hsCopied then
          AuxHashList[i].Related.State := hsRenamed
        else
          AuxHashList[i].State := hsDeleted;
    if AuxHashList[i].State = hsDeleted then
      HashList.AddCopy(AuxHashList[i]);
  end;
  // 4. Print out the results
  {$IFDEF DEBUG_OPERATIONSPROGRESS}
  DebugLog('TVerify.InternalMatch: sort new hashlist');
  {$ENDIF}
  if ShowProgress then
    DoStateChange(vstSortingResults);
  HashList.SortByPath;
  if ShowProgress then
    DoStateChange(vstWritingResults);
  for State := Low(THashState) to High(THashState) do
    StateCount[State] := 0;
  {$IFDEF DEBUG_OPERATIONSPROGRESS}
  DebugLog('TVerify.InternalMatch: print results (%d items)', [HashList.Count]);
  {$ENDIF}
  DeleteOutputFile;
  OutputStream := nil;
  try
    for i := 0 to Pred(HashList.Count) do
    begin
      {$IFDEF DEBUG_OPERATIONSPROGRESS}
      DebugLog('TVerify.InternalMatch: processing item %d', [i]);
      {$ENDIF}
      Inc(StateCount[HashList[i].State]);
      if not (HashList[i].State in HideStates) then
      begin
        if (Result = EXITCODE_OK) or (Result = EXITCODE_NO_DATA) then
          Result := EXITCODE_CHANGED_FILES;
        if OutputStream = nil then
        begin
          if OutputFileName <> '' then
          begin
            {$IFDEF DEBUG_OPERATIONSPROGRESS}
            DebugLog('TVerify.InternalMatch: create output file "%s"', [OutputFileName]);
            {$ENDIF}
            OutputStream := TTextStream.Create(OutputFileName, fmCreate, fmShareDenyWrite);
          end;
        end;
        if OutputStream <> nil then
        begin
          {$IFDEF DEBUG_OPERATIONSPROGRESS}
          DebugLog('TVerify.InternalMatch: output for item %d', [i]);
          {$ENDIF}
          OutputStream.WriteLine(HashList[i].MatchOutput(True));
        end;
      end;
      if ShowProgress then
        DoStateChange(vstMatchOutput, TStateChangeHashItem.Create(HashList[i]));
    end;
  finally
    FreeAndNil(OutputStream);
  end;
  {$IFDEF DEBUG_OPERATIONSPROGRESS}
  DebugLog('TVerify.InternalMatch: print results complete');
  {$ENDIF}
  if ShowProgress then
    DoStateChange(vstMatchStats, TStateChangeStatistics.Create(StateCount[hsOK], StateCount[hsNew], StateCount[hsChanged], StateCount[hsDeleted], StateCount[hsCopied], StateCount[hsRenamed]));
  {$IFDEF DEBUG_OPERATIONSPROGRESS}
  DebugLog('TVerify.InternalMatch: done');
  {$ENDIF}
end;

function TVerify.Match: integer;
begin
  {$IFDEF DEBUG_OPERATIONSPROGRESS}
  DebugLog('TVerify.Match: start');
  {$ENDIF}
  HashList.Clear;
  AuxHashList.Clear;
  {$IFDEF DEBUG_OPERATIONSPROGRESS}
  DebugLog('TVerify.Match: deleting output file');
  {$ENDIF}
  DeleteOutputFile;
  {$IFDEF DEBUG_OPERATIONSPROGRESS}
  DebugLog('TVerify.Match: loading hashfile 1');
  {$ENDIF}
  LoadHashFile(HashesFileName, HashList);
  {$IFDEF DEBUG_OPERATIONSPROGRESS}
  DebugLog('TVerify.Match: loading hashfile 2');
  {$ENDIF}
  LoadHashFile(MatchListFileName, AuxHashList);
  Result := InternalMatch;
  {$IFDEF DEBUG_OPERATIONSPROGRESS}
  DebugLog('TVerify.Match: done');
  {$ENDIF}
end;

function TVerify.Dupes(OneFileOnly: boolean): integer;
var
  SearchList: THashList;
  //Item: THashItem;
  i, Index: integer;
  Done, Total: int64;
  State: THashState;
  StateCount: array[THashState] of integer;
  OutputStream: TTextStream;

  procedure Step;
    begin
      Inc(Done);
      HashingProgress(Self, Done, Total);
    end;

begin
  {$IFDEF DEBUG_OPERATIONSPROGRESS}
  DebugLog('TVerify.Dupes: start');
  {$ENDIF}
  Result := EXITCODE_OK;
  HashList.Clear;
  AuxHashList.Clear;
  {$IFDEF DEBUG_OPERATIONSPROGRESS}
  DebugLog('TVerify.Dupes: deleting output file');
  {$ENDIF}
  DeleteOutputFile;
  {$IFDEF DEBUG_OPERATIONSPROGRESS}
  DebugLog('TVerify.Dupes: loading hashfile');
  {$ENDIF}
  LoadHashFile(HashesFileName, HashList);
  if OneFileOnly then
    SearchList := HashList
  else
  begin
    {$IFDEF DEBUG_OPERATIONSPROGRESS}
    DebugLog('TVerify.Dupes: loading hashfile 2');
    {$ENDIF}
    SearchList := AuxHashList;
    LoadHashFile(MatchListFileName, SearchList);
  end;
  SearchList.SortByHash;
  if ShowProgress then
    DoStateChange(vstLookingForDupes);
  Total := HashList.Count;
  Done := -1;
  Step;
  for State := Low(THashState) to High(THashState) do
    StateCount[State] := 0;
  {$IFDEF DEBUG_OPERATIONSPROGRESS}
  DebugLog('TVerify.Dupes: creating output file "%s"', [OutputFileName]);
  {$ENDIF}
  if OutputFileName = '' then
    OutputStream := nil
  else
    OutputStream := TTextStream.Create(OutputFileName, fmCreate, fmShareDenyWrite);
  try
    {$IFDEF DEBUG_OPERATIONSPROGRESS}
    DebugLog('TVerify.Dupes: matching hashlists');
    {$ENDIF}
    for i := 0 to Pred(HashList.Count) do
    begin
      HashList[i].State := hsNew;
      if SearchList.FindByHash(HashList[i].Hash, HashList[i].Size, Index) then
        repeat
          if HashList[i] <> SearchList[Index] then
          begin
            HashList[i].State := hsCopied;
            HashList[i].Related := SearchList[Index];
            if OutputStream <> nil then
              OutputStream.WriteLine(HashList[i].DupesOutput(True));
            if ShowProgress then
              DoStateChange(vstDupesOutput, TStateChangeHashItem.Create(HashList[i]));
          end;
          Inc(Index);
        until (Index >= SearchList.Count) or (HashItemCompareByHash(HashList[i], SearchList[Index]) <> 0);
      Inc(StateCount[HashList[i].State]);
      Step;
    end;
    {$IFDEF DEBUG_OPERATIONSPROGRESS}
    DebugLog('TVerify.Dupes: matching hashlists complete');
    {$ENDIF}
  finally
    FreeAndNil(OutputStream);
  end;
  if ShowProgress then
    DoStateChange(vstDupesStats, TStateChangeStatistics.Create(StateCount[hsOK], StateCount[hsNew], StateCount[hsChanged], StateCount[hsDeleted], StateCount[hsCopied], StateCount[hsRenamed]));
  {$IFDEF DEBUG_OPERATIONSPROGRESS}
  DebugLog('TVerify.Dupes: done');
  {$ENDIF}
end;

function TVerify.DeleteDupes: integer;
var
  Index, NextIndex, HashIndex, DirIndex: integer;
  State: THashState;
  StateCount: array[THashState] of integer;
  Hash: AnsiString;
  Size: int64;
  OutputStream: TTextStream;
  FileNames: TStringList;
  Dir, FileName: string;
begin
  {$IFDEF DEBUG_OPERATIONSPROGRESS}
  DebugLog('TVerify.DeleteDupes: start');
  {$ENDIF}
  Result := EXITCODE_OK;
  HashList.Clear;
  {$IFDEF DEBUG_OPERATIONSPROGRESS}
  DebugLog('TVerify.DeleteDupes: deleting output file');
  {$ENDIF}
  DeleteOutputFile;
  {$IFDEF DEBUG_OPERATIONSPROGRESS}
  DebugLog('TVerify.DeleteDupes: loading hashfile');
  {$ENDIF}
  LoadHashOrChangesFile(HashesFileName, HashList);
  HashList.SortByHash;
  for HashIndex := 0 to Pred(HashList.Count) do
    HashList[HashIndex].State := hsOK;
  for State := Low(THashState) to High(THashState) do
    StateCount[State] := 0;
  if ShowProgress then
    DoStateChange(vstLookingForDupes);
  {$IFDEF DEBUG_OPERATIONSPROGRESS}
  DebugLog('TVerify.DeleteDupes: creating output file "%s"', [OutputFileName]);
  {$ENDIF}
  if OutputFileName = '' then
    OutputStream := nil
  else
    OutputStream := TTextStream.Create(OutputFileName, fmCreate, fmShareDenyWrite);
  try
    {$IFDEF DEBUG_OPERATIONSPROGRESS}
    DebugLog('TVerify.DeleteDupes: processing dupes');
    {$ENDIF}
    FileNames := TStringList.Create;
    try
      try
        Index := 0;
        while Index < HashList.Count do
        begin
          CheckControlEvent;
          Hash := HashList[Index].Hash;
          Size := HashList[Index].Size;
          NextIndex := Succ(Index);
          while NextIndex < HashList.Count do
          begin
            if (HashList[NextIndex].Hash = Hash) and (HashList[NextIndex].Size = Size) then
              Inc(NextIndex)
            else
              Break;
          end;
          if Index < Pred(NextIndex) then
          begin
            {$IFDEF DEBUG_OPERATIONSPROGRESS}
            DebugLog('TVerify.DeleteDupes: found %d duplicates for %s', [NextIndex - Index - 1, HashList[Index].FileName]);
            {$ENDIF}
            FileNames.Clear;
            for DirIndex := 0 to Pred(BaseDirectories.Count) do
            begin
              Dir := FixUpDirectory(BaseDirectories[DirIndex]);
              for HashIndex := Index to Pred(NextIndex) do
              begin
                FileName := Dir + HashList[HashIndex].FileName;
                if FileExists(FileName) then
                  FileNames.Add(FileName);
              end;
            end;
            if (FileNames.Count > 1) and Assigned(OnDeleteDuplicates) then
            begin
              OnDeleteDuplicates(Self, FileNames);
              for HashIndex := Index to Pred(NextIndex) do
              begin
                HashList[HashIndex].State := hsDeleted;
                for DirIndex := 0 to Pred(BaseDirectories.Count) do
                begin
                  Dir := FixUpDirectory(BaseDirectories[DirIndex]);
                  FileName := Dir + HashList[HashIndex].FileName;
                  if FileExists(FileName) then
                  begin
                    HashList[HashIndex].State := hsOK;
                    Break;
                  end;
                end;
              end;
              for HashIndex := Index to Pred(NextIndex) do
              begin
                if HashList[HashIndex].State <> hsOK then
                  Result := EXITCODE_CHANGED_FILES;
                if not (HashList[HashIndex].State in HideStates) then
                begin
                  if OutputStream <> nil then
                  begin
                    OutputStream.WriteLine(HashList[HashIndex].MatchOutput(True));
                    OutputStream.Flush;
                  end;
                end;
              end;
            end;
          end;
          Index := NextIndex;
        end;
      except
        on E: EDelayedAbort do
          Result := EXITCODE_BREAK
      end;
    finally
      FreeAndNil(FileNames);
    end;
    {$IFDEF DEBUG_OPERATIONSPROGRESS}
    DebugLog('TVerify.DeleteDupes: processing dupes complete');
    {$ENDIF}
    HashList.SortByPath;
    FreeAndNil(OutputStream);
    if OutputFileName = '' then
      OutputStream := nil
    else
      OutputStream := TTextStream.Create(OutputFileName, fmCreate, fmShareDenyWrite);
    DeleteOutputFile;
    for HashIndex := 0 to Pred(HashList.Count) do
    begin
      if not (HashList[HashIndex].State in HideStates) then
      begin
        if OutputStream <> nil then
          OutputStream.WriteLine(HashList[HashIndex].MatchOutput(True));
        if ShowProgress then
          DoStateChange(vstMatchOutput, TStateChangeHashItem.Create(HashList[HashIndex]));
      end;
      Inc(StateCount[HashList[HashIndex].State]);
    end;
  finally
    FreeAndNil(OutputStream);
  end;
  if ShowProgress then
    DoStateChange(vstDupesStats, TStateChangeStatistics.Create(StateCount[hsOK], StateCount[hsNew], StateCount[hsChanged], StateCount[hsDeleted], StateCount[hsCopied], StateCount[hsRenamed]));
  {$IFDEF DEBUG_OPERATIONSPROGRESS}
  DebugLog('TVerify.DeleteDupes: done');
  {$ENDIF}
end;

function TVerify.Sort: integer;
begin
  {$IFDEF DEBUG_OPERATIONSPROGRESS}
  DebugLog('TVerify.Sort: start');
  {$ENDIF}
  //Result := EXITCODE_NO_DATA;
  HashList.Clear;
  {$IFDEF DEBUG_OPERATIONSPROGRESS}
  DebugLog('TVerify.Sort: loading hashlist "%s"', [HashesFileName]);
  {$ENDIF}
  LoadHashFile(HashesFileName, HashList);
  if ShowProgress then
    DoStateChange(vstSortingHashlist);
  HashList.SortByPath;
  {$IFDEF DEBUG_OPERATIONSPROGRESS}
  DebugLog('TVerify.Sort: saving hashlist "%s"', [HashesFileName]);
  {$ENDIF}
  HashList.SaveToFile(HashesFileName);
  Result := EXITCODE_OK;
  {$IFDEF DEBUG_OPERATIONSPROGRESS}
  DebugLog('TVerify.Sort: done');
  {$ENDIF}
end;

function TVerify.Test: integer;
begin
  {$IFDEF DEBUG_OPERATIONSPROGRESS}
  DebugLog('TVerify.Test: start');
  {$ENDIF}
  //Result := EXITCODE_NO_DATA;
  HashList.Clear;
  {$IFDEF DEBUG_OPERATIONSPROGRESS}
  DebugLog('TVerify.Test: loading hashlist "%s"', [HashesFileName]);
  {$ENDIF}
  LoadHashFile(HashesFileName, HashList);
  Result := EXITCODE_OK;
  {$IFDEF DEBUG_OPERATIONSPROGRESS}
  DebugLog('TVerify.Test: done');
  {$ENDIF}
end;

{$IFDEF OVERRIDEWOW64}
procedure TVerify.ChangeWow64Redirection(Active: boolean);
var
  OK: boolean;
begin
  if not Is64BitWindows then
    Exit;
  if Active then
    if fWow64RedirectionOldValueDONOTCHANGE <> nil then
      OK := Wow64RevertWow64FsRedirection(fWow64RedirectionOldValueDONOTCHANGE)
    else
      OK := False
  else
    OK := Wow64DisableWow64FsRedirection(fWow64RedirectionOldValueDONOTCHANGE);
  if not OK then
    Raise EOSError.CreateFmt('Failed to change WoW64 redirection state to %d.', [Integer(Active)]);
end;
{$ENDIF}

{$IFDEF RAWFILEACCESS}
procedure TVerify.ChangeRawFileAccessStatus(Active: boolean);
{$IFDEF RAWFILEACCESSUSING7ZIP}
var
  Directories: array of string;
  Directory: integer;
  FileNames: array of string;
  FileName: Integer;
{$ENDIF}
begin
  uHashItem.RawFileAccessActive := False;
  {$IFDEF RAWFILEACCESSUSING7ZIP}
  if Active <> SevenZip.IsLibraryActive then
  begin
    if Active then
    begin
      SetLength(Directories, 2);
      Directories[0] := ExtractFilePath(ParamStr(0));
      Directories[1] := '';
      SetLength(FileNames, 3);
      FileNames[0] := 'vfy' + {$IFDEF WIN64} '64' {$ELSE} '32' {$ENDIF} + '-7z.dll';
      FileNames[1] := '7z-' + {$IFDEF WIN64} '64' {$ELSE} '32' {$ENDIF} + '.dll';
      FileNames[2] := '7z.dll';
      for Directory := Low(Directories) to High(Directories) do
        for FileName := Low(FileNames) to High(FileNames) do
          if SevenZip.InitLibrary(Directories[Directory] + FileNames[FileName]) then
          begin
            uHashItem.RawFileAccessActive := True;
            Break;
          end;
    end
    else
    begin
      SevenZip.DoneLibrary;
    end;
  end;
  {$ENDIF}
end;
{$ENDIF}

{$IFDEF CASESENSITIVESUPPORT}
{$IFDEF DELPHI6_UP}
  {$WARN UNIT_PLATFORM OFF}
  {$WARN SYMBOL_PLATFORM OFF}
{$ENDIF}

{$IFNDEF DELPHI6_UP}
  // Delphi 5 has an incorrect declaration of Windows.FindFirstFileEx
  function FindFirstFileEx(lpFileName: PChar; fInfoLevelId: TFindexInfoLevels;
    lpFindFileData: Pointer; fSearchOp: TFindexSearchOps; lpSearchFilter: Pointer;
    dwAdditionalFlags: DWORD): THandle; stdcall; external kernel32 name 'FindFirstFileExA';
{$ENDIF}

  // Copied from SysUtils
  function FindMatchingFile(var F: TSearchRec): Integer;
  var
    LocalFileTime: TFileTime;
  begin
    with F do
    begin
      while FindData.dwFileAttributes and ExcludeAttr <> 0 do
        if not FindNextFile(FindHandle, FindData) then
        begin
          Result := GetLastError;
          Exit;
        end;
      FileTimeToLocalFileTime(FindData.ftLastWriteTime, LocalFileTime);
      FileTimeToDosDateTime(LocalFileTime, LongRec(Time).Hi,
        LongRec(Time).Lo);
      Size := FindData.nFileSizeLow;
      Attr := FindData.dwFileAttributes;
      Name := FindData.cFileName;
    end;
    Result := 0;
  end;

function TVerify.FindFirst(const Path: string; Attr: Integer; var SR: TSearchRec): Integer;
begin
  // This will only set-up the contents of SR
  Result := SysUtils.FindFirst(Path, Attr, SR);
  if CaseSensitive then
  begin
    SysUtils.FindClose(SR);
    // Now I will create the search handle with required parameters
    SR.FindHandle := FindFirstFileEx(PChar(Path), FindExInfoStandard, @SR.FindData, FindExSearchNameMatch, nil, FIND_FIRST_EX_CASE_SENSITIVE);
    if SR.FindHandle <> INVALID_HANDLE_VALUE then
    begin
      Result := FindMatchingFile(SR);
      if Result <> 0 then SysUtils.FindClose(SR);
    end
    else
      Result := GetLastError;
  end;
end;

{$IFDEF DELPHI6_UP}
  {$WARN UNIT_PLATFORM ON}
  {$WARN SYMBOL_PLATFORM ON}
{$ENDIF}
{$ENDIF}

function TVerify.ProcessFiles(const BasePath, Path: string; ProcessFileFunction: TProcessFileFunction; Tag: integer): integer;
const
  ITEM_IS_FILE = 0;
  ITEM_IS_DIRECTORY = 1;
var
  SR: TSearchRec;
  List: TStringList;
  i, Code: integer;
begin
  List := TStringList.Create;
  try
    if FindFirst(BasePath + Path + '*.*', faAnyFile, SR) = 0 then
      try
        repeat
          CheckControlEvent;
          {$IFDEF DELPHI7_UP}
            {$WARN SYMBOL_PLATFORM OFF}
          {$ENDIF}
          if (not Junctions) and Longbool(SR.FindData.dwFileAttributes and FILE_ATTRIBUTE_REPARSE_POINT) then
          begin
            // Skip
          end
          else
          if Longbool(SR.Attr and faDirectory) then
          begin
            if Recursive and (SR.Name <> '.') and (SR.Name <> '..') then
              List.AddObject(SR.Name, TObject(ITEM_IS_DIRECTORY));
          end
          {$IFNDEF DELPHI2007_UP}
          else if Longbool(SR.Attr and faVolumeID) then
          begin
            // Nothing to do
          end
          {$ENDIF}
          {$IFDEF DELPHI7_UP}
            {$WARN SYMBOL_PLATFORM ON}
          {$ENDIF}
          else
          begin
            List.AddObject(SR.Name, TObject(ITEM_IS_FILE));
          end;
        until FindNext(SR) <> 0;
      finally
        SysUtils.FindClose(SR);
      end;
    List.Sort;
    Result := EXITCODE_OK;
    for i := 0 to Pred(List.Count) do
    begin
      CheckControlEvent;
      Code := EXITCODE_OK;
      if List.Objects[i] = TObject(ITEM_IS_DIRECTORY) then
        Code := ProcessFiles(BasePath, Path + List[i] + '\', ProcessFileFunction, Tag)
      else if IsFileWanted(Path + List[i]) then
      begin
        if ShowProgress then
          DoStateChange(vstProcessingFile, TStateChangeFileName.Create(Path + List[i]));
        Code := ProcessFileFunction(BasePath, Path + List[i], Tag);
        if ShowProgress then
          DoStateChange(vstProgressDone);
      end;
      if Code <> EXITCODE_OK then
        Result := Code;
    end;
  finally
    FreeAndNil(List);
  end;
end;

procedure TVerify.HashingProgress(Sender: TObject; Done, Total: int64);
begin
  CheckControlEvent;
  if ShowProgress then
    DoProgress(Done, Total);
end;

function CreateRegEx(const Pattern: string): TRegExp;
begin
  Result := RegExCreate(Pattern, [rcoSingleLine, rcoIgnoreCase]);
end;

function CreateRegExFromFileMask(Mask: string): TRegExp;
var
  i: integer;
  Pattern: string;
begin
  Mask := ExtractFileName(Mask); // Ignore path
  Pattern := '';
  for i := 1 to Length(Mask) do
    case Mask[i] of
      '*':
        Pattern := Pattern + '.*';
      '?':
        Pattern := Pattern + '.';
      ']':
        Pattern := Pattern + '\]';
      '^':
        Pattern := Pattern + '\^';
      '\':
        Pattern := Pattern + '\\';
      else
        Pattern := Pattern + '[' + Mask[i] + ']';
    end;
  Result := CreateRegEx('^' + Pattern + '$');
end;

procedure TVerify.ClearIncludeExcludeMasks;
var
  i: integer;
begin
  for i := 0 to Pred(IncludeExcludeList.Count) do
    RegExFree(TRegExp(IncludeExcludeList.Objects[i]));
  IncludeExcludeList.Clear;
end;

procedure TVerify.ClearProgressFile;
begin
  if ProgressFileName <> '' then
    if FileExists(ProgressFileName) then
      SysUtils.DeleteFile(ProgressFileName);
end;

procedure TVerify.AddIncludeExcludeMask(const Mask: string; MaskType: TIncludeExcludeMask);
begin
  if IncludeExcludeList.Count = 0 then
  begin
    case MaskType of
      iemInclude, iemIncludeRegExp:
        IncludeExcludeList.AddObject(INCEXCTYPE_EXCLUDE, CreateRegEx('.*'));
      iemExclude, iemExcludeRegExp:
        IncludeExcludeList.AddObject(INCEXCTYPE_INCLUDE, CreateRegEx('.*'));
    end;
  end;
  case MaskType of
    iemInclude:
      IncludeExcludeList.AddObject(INCEXCTYPE_INCLUDE + Mask, CreateRegExFromFileMask(Mask));
    iemExclude:
      IncludeExcludeList.AddObject(INCEXCTYPE_EXCLUDE + Mask, CreateRegExFromFileMask(Mask));
    iemIncludeRegExp:
      IncludeExcludeList.AddObject(INCEXCTYPE_INCLUDE + Mask, CreateRegEx(Mask));
    iemExcludeRegExp:
      IncludeExcludeList.AddObject(INCEXCTYPE_EXCLUDE + Mask, CreateRegEx(Mask));
  end;
end;

procedure TVerify.ApplyIncludeExcludeToHashList(HashList: THashList);
var
  i: integer;
begin
  for i := Pred(HashList.Count) downto 0 do
    if not IsFileWanted(HashList[i].Path) then
      HashList.Delete(i);
end;

function TVerify.IsFileWanted(const Path: string): boolean;
var
  i: integer;
  s: string;
begin
  Result := True;
  for i := 0 to Pred(IncludeExcludeList.Count) do
    if TRegExp(IncludeExcludeList.Objects[i]).Match(Path) then
    begin
      s := IncludeExcludeList[i];
      if (s <> '') and (s[1] = INCEXCTYPE_INCLUDE) then
        Result := True
      else if (s <> '') and (s[1] = INCEXCTYPE_EXCLUDE) then
        Result := False;
    end;
end;

procedure TVerify.CheckControlEvent;
begin
  if Aborted then
  begin
    {$IFDEF DEBUG_OPERATIONSPROGRESS}
    DebugLog('TVerify.CheckControlEvent: task aborted');
    {$ENDIF}
    Aborted := False;
    Raise EDelayedAbort.Create('Task aborted.');
  end;
end;

{$IFDEF DEBUG_LOG}
procedure TVerify.DebugLog(const Msg: string);
begin
  if LogStream <> nil then
  begin
    try
      LogStream.WriteLine(FormatDateTime('yyyy"-"mm"-"dd" "hh":"nn":"ss"."zzz": "', Now) + Msg);
    except
      // Drop any errors
    end;
  end;
end;

procedure TVerify.DebugLog(const Msg: string; const Params: array of const);
var
  s: string;
begin
  try
    s := Format(Msg, Params);
  except
    s := 'Invalid DebugLog format: ' + Msg;
  end;
  DebugLog(s);
end;

procedure TVerify.SetLogFileName(const Value: string);
begin
  FreeAndNil(fLogStream);
  fLogFileName := '';
  if Value <> '' then
  begin
    fLogStream := TTextStream.Create(Value, fmCreate, fmShareDenyWrite);
    fLogFileName := Value;
  end;
end;
{$ENDIF}

end.
